#ifndef DICTIONARY_HH
#define DICTIONARY_HH

#include <vector>
#include <exception>
#include <cassert>  // only to check if unsigned < int_max
#include <climits>  // only to check if unsigned < int_max
#include <typeinfo> // only needed in sizeInBytes

/*
 * Implements a dictionary based on an open-addressing hash table. 
 * Used to apply dictionary compression of data of type Tdata.
 * Maps Tdata to unsigned.
 */
template <typename Tdata, typename Thashfunc>
class Dictionary {
  public:
    Dictionary (unsigned aNoEntries, Thashfunc aFunc, bool aCopyValues);
    ~Dictionary();
  public:
    std::vector<unsigned> encode (const std::vector<Tdata>&    aColumn);
    unsigned              encode (const Tdata&                 aColumnValue);
    std::vector<Tdata>    decode (const std::vector<unsigned>& aCompressedColumn);
    Tdata                 decode (const unsigned               aCompressedColumnValue);
  public:
    unsigned sizeInBytes ();
  private:
    unsigned insert (const Tdata& aVal);
    int      search (const Tdata& aVal);
  private:
    std::vector<const Tdata*> _data;
    Thashfunc                 _hashfunc;
    bool                      _copyValues;
};


template <typename Tdata, typename Thashfunc>
Dictionary<Tdata, Thashfunc>::Dictionary (unsigned aNoEntries, Thashfunc aFunc, bool aCopyValues)
  : _data(aNoEntries, NULL), _hashfunc(aFunc), _copyValues(aCopyValues)
  {}


template <typename Tdata, typename Thashfunc>
Dictionary<Tdata, Thashfunc>::~Dictionary () {
  for (const Tdata* elem : _data) {
    if (elem != NULL) {
      delete elem;
    }
  }
}


template <typename Tdata, typename Thashfunc>
std::vector<unsigned> Dictionary<Tdata, Thashfunc>::encode (const std::vector<Tdata>& aColumn) {
  std::vector<unsigned> compressedColumn;
  for (const auto& val : aColumn) {
    compressedColumn.push_back( encode(val) );
  }
  return compressedColumn;
}


template <typename Tdata, typename Thashfunc>
unsigned Dictionary<Tdata, Thashfunc>::encode (const Tdata& aColumnValue) {
  int idx = search(aColumnValue);
  if (idx >= 0) {
    return idx;
  } else {
    return insert(aColumnValue);
  }
}


template <typename Tdata, typename Thashfunc>
std::vector<Tdata> Dictionary<Tdata, Thashfunc>::decode (const std::vector<unsigned>& aCompressedColumn) {
  std::vector<Tdata> column;
  for (const auto& val : aCompressedColumn) {
    column.push_back( decode(val) );
  }
  return column;
}


template <typename Tdata, typename Thashfunc>
Tdata Dictionary<Tdata, Thashfunc>::decode (const unsigned aCompressedColumnValue) {
  assert (aCompressedColumnValue < _data.size());
  if (aCompressedColumnValue >= _data.size() || _data[aCompressedColumnValue] == NULL) {
    throw std::invalid_argument("Unknown compressed value.");
  }
  Tdata columnValue = *(_data[aCompressedColumnValue]);
  return columnValue;
}


template <typename Tdata, typename Thashfunc>
unsigned Dictionary<Tdata, Thashfunc>::insert (const Tdata& aVal) {
  unsigned i = 0;
  do {
    unsigned idx = _hashfunc(aVal, i);
    idx = idx % _data.size(); // safe guard
    if (_data[idx] == NULL) {
      const Tdata* valInsert;
      if (_copyValues) {
        valInsert = new Tdata(aVal);
      } else {
        valInsert = &aVal;
      }
      _data[idx] = valInsert;
      return idx;
    } else {
      ++i;
    }
  } while (i < _data.size());
  throw std::overflow_error("Hash table overflow.");
}


template <typename Tdata, typename Thashfunc>
int Dictionary<Tdata, Thashfunc>::search (const Tdata& aVal) {
  unsigned i = 0;
  unsigned idx;
  do {
    idx = _hashfunc(aVal, i);
    idx = idx % _data.size(); // safe guard
    if ( _data[idx] != NULL && *(_data[idx]) == aVal ) {
      assert (idx <= INT_MAX);
      return (int)idx;
    }
    ++i;
  } while (_data[idx] != NULL && i < _data.size());
  return -1;
}


template <typename Tdata, typename Thashfunc>
unsigned Dictionary<Tdata, Thashfunc>::sizeInBytes () {
  unsigned size = 0;
  size += _data.size() * sizeof(std::string*);

  if (typeid(Tdata) == typeid(std::string)) {
    for (const std::string* stringval : _data) {
      if (stringval != NULL) {
        size += sizeof(std::string);
        size += stringval->size() * sizeof(char);
      }
    }
  } else {
    std::cout << __FILE__ << ":" << __LINE__
              << ": WARNING: This function may not do what you intend." << std::endl;
    for (const Tdata* elem : _data) {
      if (elem != NULL) {
        size += sizeof(*elem);
      }
    }
  }

  return size;
}

#endif
