#include <vector>
#include <fstream>
#include <iostream>
#include <iomanip>

#include "dictionary.hh"

/*
 * This file illustrates the benefits of dicionary compression.
 * In the main function, the following steps are executed:
 *   - Read in a column of data from file.
 *   - Create a simple statistical profile of the data.
 *   - Create a dictionary. The dictionary is implemented using
 *     open addressing, hence, it requires hash functions that take an additional
 *     parameter, that is, the probing number.
 *   - Encode all data using the dictionary.
 *   - Compare the size of the uncompressed data with the size of the
 *     compressed data.
 */


/*
 * hash function extended by probing number
 * cf. Introduction to Algorithms, Chap. 11.4: Open addressing
 * Note: Only LinearProbingHash will be used.
 */
template <typename Tdata, typename TauxHashFunc>
struct LinearProbingHash {
  unsigned operator() (const Tdata& aVal, unsigned aProbeNumber) {
    return (h(aVal) + aProbeNumber) % m;
  }
  TauxHashFunc h;
  unsigned     m;
};


template <typename Tdata, typename TauxHashFunc>
struct QuadraticProbingHash {
  unsigned operator() (const Tdata& aVal, unsigned aProbeNumber) {
    return ((unsigned)(h(aVal)  +  c1 * aProbeNumber + c2 * (aProbeNumber*aProbeNumber) )) % m;
  }
  TauxHashFunc h;
  double c1, c2;
  unsigned     m;
};


template <typename Tdata, typename TauxHashFunc1, typename TauxHashFunc2>
struct DoubleHashing {
  unsigned operator() (const Tdata& aVal, unsigned aProbeNumber) {
    return ( h1(aVal)  +  aProbeNumber * h2(aVal) ) % m;
  }
  TauxHashFunc1 h1;
  TauxHashFunc2 h2;
  unsigned     m;
};


// ordinary hash function
unsigned sumStringHash (const std::string& word) {
  unsigned sum = 0;
  for (unsigned i = 0; i < word.size(); ++i) {
    sum += word[i];
  }
  return sum;
}


// other ordinary hash function
struct LarsonStringHash {
  unsigned operator() (std::string word) {
    unsigned h = salt;
    const char* s = word.data(); 
    // data() returns a pointer to an array that contains 
    // the same sequence of characters as the characters 
    // that make up the value of the string object.
    
    while (*s) {
      h = h * 101 + *s++;
    }
    return h;
  }

  unsigned salt;
};


// statistical profile of an attribute
// will be used to store information about country_or_area as given in country_or_area.csv
struct StatisticalProfile {
  std::string attributeName;
  unsigned    cardinality;
  unsigned    noDistinctValues;
};


// read the attribute values of an attribute from a file
// will be used to read in the values of country_or_area as given in country_or_area.csv
std::vector<std::string> readColumn (std::string fileDescriptor) {
  std::vector<std::string> countries;
  std::ifstream columnData (fileDescriptor);
  std::string country;
  while (std::getline(columnData, country)) {
    countries.push_back(country);
  }
  return countries;
}


int main () {
  // read in attribute values
  std::vector<std::string> countries = readColumn ("country_or_area.csv");

  // prepare statistical profile of attribute values
  StatisticalProfile countryProfile;
  countryProfile.attributeName    = "country_or_area";
  countryProfile.cardinality      = 884244;
  countryProfile.noDistinctValues = 238;

  // hash function for open addressing hash table
  // needed by Dictionary class
  using sumString_f = unsigned (*) (const std::string&);
  using LinearProbingString_t = LinearProbingHash<std::string, sumString_f>;
  LinearProbingString_t linHash;
  linHash.h = sumStringHash;
  linHash.m = 2*countryProfile.noDistinctValues;

  using QuadraticProbingString_t = QuadraticProbingHash<std::string, sumString_f>;
  QuadraticProbingString_t quadHash;
  quadHash.h = sumStringHash;
  quadHash.m = (1<<10);
  quadHash.c1 = 0.5;
  quadHash.c2 = 0.5;

  using DoubleHashingString_t = DoubleHashing<std::string, sumString_f, LarsonStringHash>;
  DoubleHashingString_t doubleHash;
  doubleHash.h1 = sumStringHash;
  doubleHash.h2 = LarsonStringHash{111};
  doubleHash.m = 2*countryProfile.noDistinctValues;

  // Dictionary
  //Dictionary<std::string, LinearProbingString_t> dict (2*countryProfile.noDistinctValues, linHash, true);
  //Dictionary<std::string, QuadraticProbingString_t> dict (2*countryProfile.noDistinctValues, quadHash, true);
  Dictionary<std::string, DoubleHashingString_t> dict (2*countryProfile.noDistinctValues, doubleHash, true);
  std::vector<unsigned> encoding = dict.encode(countries);
  std::vector<std::string> decoding = dict.decode(encoding);

  bool print = false;
  if (print) {
    std::cout << "countries.size(): " << countries.size() << std::endl;

    std::cout << "encoding.size(): " << encoding.size() << std::endl;
    for (unsigned i = 0; i < encoding.size(); ++i) {
      std::cout << i << ": " << encoding[i] << std::endl;;
    }

    std::cout << "decoding.size(): " << decoding.size() << std::endl;
    for (unsigned i = 0; i < decoding.size(); ++i) {
      std::cout << i << ": " << decoding[i] << std::endl;
    }
  }

  unsigned sizeUncompressedCountries = countries.size() * sizeof(std::string);
  for (const std::string& val : countries) {
    sizeUncompressedCountries += val.size() * sizeof(char);
  }

  unsigned sizeCompressedCountries = encoding.size() * sizeof(unsigned);
  unsigned sizeDictionary = dict.sizeInBytes();
  unsigned totalSizeCompressed = sizeCompressedCountries + sizeDictionary;

  std::cout << "Size in bytes of uncompressed data:     " << std::setw(10) << sizeUncompressedCountries << std::endl;
  std::cout << "Size in bytes of compressed data:       " << std::setw(10) << sizeCompressedCountries << std::endl;
  std::cout << "Size in bytes of dictionary:            " << std::setw(10) << sizeDictionary << std::endl;
  std::cout << "Size of compressed data and dictionary: " << std::setw(10) << totalSizeCompressed << std::endl; 
  std::cout << "Compression ratio:        " << ( (double) sizeUncompressedCountries / totalSizeCompressed) << std::endl;
  std::cout << "Percentage space savings: " << (1 - (double) totalSizeCompressed / sizeUncompressedCountries) << std::endl;
}
